/****************************************************************************
 * WangLandau: WangLandau simulation for RNA turner energy's density of state
 * Copyright (C) 2008-2009 Laboratoire de Recherche Informatique (LRI)
 * All Rights Reserved
 *
 *     This source code is distributed under the terms of the
 *     GNU General Public License.
 ****************************************************************************/

/*
 * main.c
 * main wangLandau simulation file
 * 2009/06/04 by Feng LOU & Peter CLOTE
 */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <limits.h>
#include <string.h>
#include <time.h>
#include <signal.h>
#include <errno.h>
#include <unistd.h>
#include "../lib/rnaSecStr.h"
#include "../lib/tools.h"
#include "../lib/struct.h"
#include "../lib/main.h"
#include "../ext/lib/fold.h"
#include "../ext/lib/fold_vars.h"
#include "../ext/lib/part_func.h"
#include "../ext/lib/cofold.h"

// keyboard listener variable
#define PROMPT "Writing the temporary results into file ......"
char *prompt = PROMPT;
static int IS_CTRL_C = 0;
// Debug variable
static int DEBUG = 0;
static int FINE_DEBUG = 0;
// Program Option variable
static double FLATNESS_CONSTANT = 0.8;
static int MC_step = 100;
static int SIZE = NMAX;
static double DS_INIT;
static double epsilon = 0.001;
static int SS_CHANGE = 0;
static int STRUCT_PROBA = 0;
static int TOTAL_STEPS = 0;
static char* FILE_NAME = "out";
static char* SS_INITIAL;

void ctrl_c_op(int signo) {
	printf("\nProgram stop by Ctrl-C\n");
	write(STDERR_FILENO, prompt, strlen(prompt));
	printf("\n\n");
	IS_CTRL_C = 1;
}

char *tokenize(char *line) {
	char *token, *copy, *ctmp;
	int cut = -1;

	copy = (char *) space(strlen(line) + 1);
	ctmp = (char *) space(strlen(line) + 1);
	(void) sscanf(line, "%s", copy);
	ctmp[0] = '\0';
	token = strtok(copy, "&");
	cut = strlen(token) + 1;
	while (token) {
		strcat(ctmp, token);
		token = strtok(NULL, "&");
	}
	if (cut > strlen(ctmp))
		cut = -1;
	if (cut > -1) {
		if (cut_point == -1)
			cut_point = cut;
		else if (cut_point != cut) {
			fprintf(stderr,"cut_point = %d cut = %d\n", cut_point, cut);
			nrerror("Sequence and Structure have different cut points.");
		}
	}
	free(copy);

	return ctmp;
}

/*
 * return list of all base pair can be added to the secondary structure: secStr
 */
list* listOfNextStatesByAddingBasePair(char* rna, char* secStr, list* L) {
	int k, t;
	list* basePairsWhichCanBeAdded = init_list();
	int n = strlen(rna);
	for (k = 0; k < n; k++) {
		if (secStr[k] == '.') {
			for (t = k + 1; t < n; t++) {
				if (secStr[t] == '.') {
					// test THRESHOLD
					if (t - k > THRESHOLD) {
						if (basePair(rna[k], rna[t]) && canAddBasePair(k, t, n,
								L)) {
							add_list_item(basePairsWhichCanBeAdded, k, t);
						}
					}
					// test Hydrydization if two sequence with cut_point
					else {
						int cp = cut_point - 1;
						if (k < cp && t >= cp) {
							if (basePair(rna[k], rna[t]) && canAddBasePair(k,
									t, n, L)) {
								add_list_item(basePairsWhichCanBeAdded, k, t);
							}
						}
					}
				}
			}

		}
	}

	return basePairsWhichCanBeAdded;
}
/*
 * return list of all base pair can be deleted from the secondary structure: secStr
 */
list* listOfNextStatesByRemovingBasePair(char* rna, char* secStr, list* L) {
	return L;
}
/*
 * check if the table of histogram of energy turner is flat or not
 * flat: if all used item in histogram >= FLATNESS_CONSTANT * (sum_all_item/nb_item)
 */
int flat(int* h, int size) {
	int i, num;
	// if the histogram is empty
	if (h[SIZE - 1] == 0 || size == 0)
		return 0;
	// count the number of case no empty
	num = 0;
	for (i = 0; i < size; i++) {
		if (h[i] > 0)
			num++;
	}
	// check each case no empty
	for (i = 0; i < size; i++) {
		if (h[i] > 0 && h[i] < FLATNESS_CONSTANT * h[SIZE - 1] / num)
			return 0;
	}
	return 1;
}

/*		int, int**, list_ss*);
 *
 * return the table of relative density of states after the normalization
 */
double* normalizedDensityOfStates(double* g, int size) {
	int i;
	double sum = 0.0;
	for (i = 0; i < size; i++) {
		sum = sum + g[i];
	}
	for (i = 0; i < size; i++) {
		g[i] = g[i] / sum;
	}
	if (sum != 0)
		DS_INIT = DS_INIT / sum;
	return g;
}
/*
 * decide to go to another neighbor state or stay in the same state
 */
configuration nextConfiguration(char* rna, configuration cf, double c,
		double energy_min, int repet_i, list_ss* lss) {
	// calculate the number of base pair can be add or delete
	list* L0 = listOfNextStatesByAddingBasePair(rna, cf.secStr, cf.L);
	list* L1 = listOfNextStatesByRemovingBasePair(rna, cf.secStr, cf.L);
	int m0 = L0->size;
	int m1 = L1->size;
	// copy the secStr and List
	char* tmp = copyTo_char(cf.secStr);
	list* tmpList = copy_list(cf.L);
	// get the random value between 0 and (m0 + m1)
	int randValue = (int) ((float) rand() / (RAND_MAX) * (m0 + m1));
	int x = randValue % (m0 + m1);
	// get the next step maybe to go
	int i, p, q;
	if (x < m0) {
		p = get_list_x(L0, x);
		q = get_list_y(L0, x);
		tmp[p] = '(';
		tmp[q] = ')';
		add_list_item(tmpList, p, q);
	} else {
		p = get_list_x(L1, x - m0);
		q = get_list_y(L1, x - m0);
		tmp[p] = '.';
		tmp[q] = '.';
		remove_list_item(tmpList, x - m0);
	}
	if (FINE_DEBUG == 1)
		printSecStrListAsString(tmp);
	// free memory space for L0
	free_list(L0);
	// calculate turner energy
	float E1 = energy_of_struct(rna, tmp);
	// in interval 0.1
	double y0 = round(10*cf.energy) / 10.0;
	double y1 = round(10*E1) / 10.0;
	// index in the table h and g
	energy_min = energy_min / 1.0;
	int x0 = round((y0 - energy_min)/0.1);
	int x1 = round((y1 - energy_min)/0.1);
	if (FINE_DEBUG == 1)
		printf("%g %g, %g, %d, %d\n", energy_min, cf.energy, E1, x0, x1);
	if (repet_i % 1000 == 0)
		cf.g = normalizedDensityOfStates(cf.g, cf.size);
	// get the random value between 0 and 1
	double random = (double) rand();
	double uniRandValue = random / (double) RAND_MAX;
	// move to the neighborhood state by WangLandau's condition
	if (cf.g[x1] == 0 || uniRandValue < (double) (cf.g[x0] / cf.g[x1])) {
		if (x1 >= cf.size && x1 < SIZE - 1) {
			cf.size = x1 + 1;
		}
		// if the size of table cf.h and cf.g are not large enough,
		// we will create a new table with SIZE = 2 * OLD_SIZE and copy the old table inside
		while (x1 >= SIZE - 1) { // SIZE = 2*SIZE;
			cf.h = new_h(cf.h, SIZE);
			cf.g = new_g(cf.g, SIZE, DS_INIT);
			cf.lss = new_lss(cf.lss, SIZE);
			SIZE = 2 * SIZE;
		}
		cf.flag = 1;
		free(cf.secStr);
		cf.secStr = tmp;
		static char* SS_INITIAL;
		free_list(cf.L);
		cf.L = tmpList;
		cf.energy = E1;
		cf.h[x1]++;
		cf.h[SIZE - 1]++;
		cf.g[x1] = cf.g[x1] * c;
		if (cf.lss[x1] == NULL) {
			cf.lss[x1] = copyTo_char(tmp);
		}
	}
	// stay in the same state
	else {
		if (x0 >= cf.size && x0 < SIZE - 1) {
			cf.size = x0 + 1;
		}
		// if the size of table cf.h and cf.g are not large enough,
		// we will create a new table with SIZE = 2 * OLD_SIZE and copy the old table inside
		while (x0 >= SIZE - 1) { // SIZE = 2*SIZE;
			cf.h = new_h(cf.h, SIZE);
			cf.g = new_g(cf.g, SIZE, DS_INIT);
			cf.lss = new_lss(cf.lss, SIZE);
			SIZE = 2 * SIZE;
		}
		cf.flag = 0;
		cf.h[x0]++;
		cf.h[SIZE - 1]++;
		cf.g[x0] = cf.g[x0] * c;
		if (cf.lss[x0] == NULL) {// Global Statistic variable
			static char* SS_INITIAL;
			cf.lss[x0] = copyTo_char(cf.secStr);
		}
		free(tmp);
		free_list(tmpList);
	}
	return cf;
}
/*
 *	principal function of wanglandau simulation
 *	return with the density of state and the histogram of energy turner
 */
wl wangLandau(char* rna, double energy_min) {
	int i, j, size, sr; // size is the number of the case in the table h (histogram) and g (density of state)
	int num = 0;
	double c = exp(1.0);
	// initialization a new configuration
	configuration confg;
	DS_INIT = (double) (1.0 / SIZE);
	// initialization confg.g and confg.size
	confg.g = (double*) malloc(SIZE * sizeof(double));
	for (i = 0; i < SIZE; i++) {
		confg.g[i] = DS_INIT;
	}
	confg.size = 0;
	// initialization confg.secStr and confg.L
	if (SS_CHANGE == 1) {
		confg.secStr = getInitialSecStr(SS_INITIAL);
		confg.L = set_list(confg.secStr);
	} else {
		confg.secStr = initialEmptySecStr(strlen(rna));
		confg.L = init_list();
	}
	// initialization confg.lss
	confg.lss = (char**) malloc(SIZE * sizeof(char*));
	for (i = 0; i < SIZE; i++) {
		confg.lss[i] = NULL;
	}
	// initialization confg.h
	confg.h = (int*) malloc(SIZE * sizeof(int));
	// initialization confg.energy
	confg.energy = energy_of_struct(rna, confg.secStr);
	// initialization list of secondary structure
	list_ss* lss = init_list_ss();
	while (c > (1 + epsilon)) {
		// reset confg.h
		for (i = 0; i < SIZE; i++) {
			confg.h[i] = 0;
		}
		// break when is flat or the histogram is empty
		while (flat(confg.h, confg.size) == 0) {
			for (j = 0; j < MC_step; j++) {
				confg = nextConfiguration(rna, confg, c, energy_min, j, lss);
			}
			confg.g = normalizedDensityOfStates(confg.g, confg.size);
			num += MC_step;
			if (num > TOTAL_STEPS && TOTAL_STEPS != 0)
				break;
			// Write the partial RNA energy distribution stop by Ctrl-C
			if (IS_CTRL_C == 1) {
				char* tfn = (char *) calloc(20, sizeof(char));
				sprintf(tfn, "out/%s.ctrl-c.txt", FILE_NAME);
				FILE *tf = fopen(tfn, "w");
				fprintf(tf, "RNA sequence: %s\n", rna);
				fprintf(tf, "number of MonteCarlo moves: %d\n\n", num);
				printDensityOfStatesToFile(confg.g, confg.lss, confg.size,
						energy_min, DS_INIT, tf);
				fprintf(tf, "energy minimal is: %5.1f\n", energy_min);
				fclose(tf);
				exit(0);
			}
		}
		if (num > TOTAL_STEPS && TOTAL_STEPS != 0)
			break;
		if (DEBUG == 1) {
			printf("%d\n,", confg.flag);
			print_char(confg.secStr);
			print_char(rna);
			printf("Density of states: ");
			print_double(confg.g, DS_INIT, confg.size);
			printf("\n");
			printf("Histogram: ");
			print_int(confg.h, SIZE);
			printf("Flat\?: %d\n", flat(confg.h, confg.size));
		}
		c = sqrt(c);
		if (DEBUG == 1) {
			printf("Modification factor changed to: %f\n\n", c);
		}
	}
	wl w;
	w.num = num;
	w.g = confg.g;
	w.h = confg.h;
	w.lss = confg.lss;
	w.size = confg.size;
	// FREE
	free_list_ss(lss);
	free_list(confg.L);
	free(confg.secStr);
	return w;
}

/*
 * main function
 */
int main(int argc, char *argv[]) {
	if (argc < 3) {
		usage();
		exit(1);
	}
	int i, n, r;
	int flag = 0;
	char* rnaStringOrFile;
	char* s;
	FLATNESS_CONSTANT = 0.8;
	// read command options
	for (i = 1; i < argc; i++) {
		if (argv[i][0] == '-')
			switch (argv[i][1]) {
			case 's':
				flag = 1;
				rnaStringOrFile = argv[i + 1];
				break;
			case 'f':
				flag = 2;
				rnaStringOrFile = argv[i + 1];
				break;
			case 'n':
				MC_step = atoi(argv[i + 1]);
				break;
			case 'c':
				FLATNESS_CONSTANT = atof(argv[i + 1]);
				break;
			case 'e':
				epsilon = atof(argv[i + 1]);
				break;
			case 'i':
				SS_CHANGE = 1;
				SS_INITIAL = argv[i + 1];
				break;
			case 'o':
				FILE_NAME = argv[i + 1];
				break;
			case 'p':
				STRUCT_PROBA = atoi(argv[i + 1]);
				break;
			case 't':
				if (argv[i][2] != '\0')
					usage();
				if (i == argc - 1)
					usage();
				r = sscanf(argv[++i], "%lf", &temperature);
				if (!r)
					usage();
				break;
			case 'N':
				TOTAL_STEPS = atoi(argv[i + 1]);
				break;
			default:
				usage();
				exit(0);
			}
	}
	// keyboard listener to ctrl_c
	struct sigaction act;
	act.sa_handler = ctrl_c_op;
	sigemptyset(&act.sa_mask);
	act.sa_flags = 0;
	if (sigaction(SIGINT, &act, NULL) < 0) {
		fprintf(stderr, "Install Signal Action Error:%s\n\a", strerror(errno));
		exit(1);
	}
	// Read the RNA sequence and information
	char* seqInfo = (char*) malloc(100 * sizeof(char));
	s = getInputRNA(flag, rnaStringOrFile, seqInfo);
	n = strlen(s);
	// write the rna energy distribution into "FILE_NAME".out
	char* ED_FILE_NAME = (char*) calloc(100, sizeof(char));
	sprintf(ED_FILE_NAME, "out/%s.txt", FILE_NAME);
	FILE *f = fopen(ED_FILE_NAME, "w");
	if (f == NULL) {
		printf("Error opening %s for writing. Program terminated.", FILE_NAME);
		abort();
	}
	if (validNucleotideString(s) == 0) {
		printf("%s: Invalid nucleotide String\n", s);
		exit(0);
	}
	// calculate the minimum energy of turner using fold() from libRNA.a
	char secStrMin[1000];
	s = tokenize(s);
	double energy_min = cofold(s, secStrMin);
	char* rna = (char*) malloc((n + 1) * sizeof(char));
	for (i = 0; i < n; i++) {
		rna[i] = s[i];
	}
	rna[n] = '\0';
	//no necessary to run WL if the number of structure is 1 (empty structure)
	char* secStr_init = initialEmptySecStr(strlen(rna));
	list* L_init = init_list();
	list* L = listOfNextStatesByAddingBasePair(rna, secStr_init, L_init);
	if (L->size > 0) {
		// RUN main function wangLandau
		wl w = wangLandau(rna, energy_min);
		// Print the result
		printDensityOfStatesToFile(w.g, w.lss, w.size, energy_min, DS_INIT, f);
		free(w.g);
		free(w.h);
		for (i = 0; i < w.size; i++) {
			free(w.lss[i]);
		}
		free(w.lss);
	}
	fclose(f);
	// FREE
	free(s);
	free(seqInfo);
	free(rna);
	free(secStr_init);
	free_list(L_init);
	free_list(L);
	return 0;
}
